#! python3

import argparse
import json
import math
import re
from socket import timeout
import sys
from typing import Any
from unicodedata import numeric
import webbrowser
# from matplotlib import table
from quantiphy import Quantity
import logging
from math import inf, nan
import os
import numpy as np
import pandas as pd
from darum.log_readers import Details, readLogs
from quantiphy import Quantity
import holoviews as hv  # type: ignore
# import hvplot           # type: ignore
# from hvplot import hvPlot
from holoviews import opts
from bokeh.models.tickers import FixedTicker, CompositeTicker, BasicTicker
from bokeh.models import NumeralTickFormatter, HoverTool
from bokeh.util.compiler import TypeScript
from bokeh.settings import settings
import os
import glob
import panel as pn
from bokeh.models.widgets.tables import NumberFormatter, BooleanFormatter
from ansi2html import Ansi2HTMLConverter

log = logging.getLogger(__name__)



def smag(i) -> str:
    return f"{Quantity(i):.3}"

def dn_is_excluded(dn, exclude_list):
    for e in exclude_list:
        regex = re.compile(e)
        if regex.search(dn):
            log.debug(f"Excluding {dn}")
            return True
    return False



class NumericalTickFormatterWithLimit(NumeralTickFormatter):
    fail_min = 0

    def __init__(self, fail_min:int, **kwargs):
        super().__init__(**kwargs)
        if __name__ == "__main__":
            print("The generated webpage will only work if this tool is run from its installed script - not the .py file in the package!")
            # This is because of how Bokeh includes these strings in the webpage. There's a "require(custom/XXXnumerical...)" that seems to catch the module name, 
            # and in the "static __name__" definition further down we need to get it the same.
            # This works correctly when using the scripts installed by e.g. poetry or pipx.
        NumericalTickFormatterWithLimit.fail_min = fail_min
        NumericalTickFormatterWithLimit.__implementation__ = TypeScript(
"""
import {NumeralTickFormatter} from 'models/formatters/numeral_tick_formatter'

export class NumericalTickFormatterWithLimit extends NumeralTickFormatter {
    static __name__ = '""" + __name__ + """.NumericalTickFormatterWithLimit'
    FAIL_MIN=""" + str(int(fail_min)) + """

    doFormat(ticks: number[], _opts: {loc: number}): string[] {
        const formatted = []
        const ticks2 = super.doFormat(ticks, _opts)
        for (let i = 0; i < ticks.length; i++) {
            if (ticks[i] < this.FAIL_MIN) {
                formatted.push(ticks2[i])
            } else {
                formatted.push('FAIL/OoR')
            }
        }
        return formatted
    }
}
""")

customJS = r"""
<script type="text/javascript">
/* Plain navigation to anchors doesn't work because the anchors are in a Panel, and each Panel is a shadowDOM. So we need to navigate programmatically. (Panel issue 6156? Closed, but not helping)
This is done by clickInterceptor.
Additionally, Panel doesn't give us Python references to the shadowDOMs that will be created by Bokeh, so to find where is an anchor in container C the best option seems to be to search from JS for known C ids inside the shadowDOMs. (Panel issue 7443) 
This is done with xfind. */

/**
 * @description Locates the first matching elements on the page, even within shadow DOMs, using a complex n-depth selector.
 * Author: Roland Ross L. Hadi
 * GitHub: https://github.com/rolandhadi/shadow-dom-selector
 */
function xfind(e,t=document){const o=performQuery(e,!1,t);return console.log(`Found ${o?1:0} element(s)`),o}function xfindAll(e,t=document){const o=performQuery(e,!0,t);return console.log(`Found ${o.length} element(s)`),o}function performQuery(e,t,o){validateSelector(e);const n=e.split(">>>");let r=o;for(let e=0;e<n.length;e++){if(e===n.length-1&&t)return querySelectorAllXFind(n[e],r);if(r=querySelectorXFind(n[e],r),null===r)return console.error(`Selector ${n[e]} not found`),t?[]:null}return r}function querySelectorAllXFind(e,t=document){return queryInShadowDOM(e,!0,t)}function querySelectorXFind(e,t=document){return queryInShadowDOM(e,!1,t)}function queryInShadowDOM(e,t,o){let n=o.querySelector(e);if(document.head.createShadowRoot||document.head.attachShadow){if(!t&&n)return n;return splitByUnquotedCharacter(e,",").reduce(((e,n)=>{if(!t&&e)return e;const r=splitByUnquotedCharacter(n.trim().replace(/\s*([>+~])\s*/g,"$1")," ").filter(Boolean),l=r.length-1,u=gatherAllElementsXFind(r[l],o),s=matchElements(r,l,o);return t?e.concat(u.filter(s)):u.find(s)||null}),t?[]:null)}return t?o.querySelectorAll(e):n}function matchElements(e,t,o){return n=>{let r=t,l=n;for(;l;){const t=!!l.matches&&l.matches(e[r]);if(t&&0===r)return!0;t&&r--,l=getParentOrHost(l,o)}return!1}}function splitByUnquotedCharacter(e,t){return e.match(/\\?.|^$/g).reduce(((e,o)=>('"'!==o||e.singleQuote?"'"!==o||e.doubleQuote?e.doubleQuote||e.singleQuote||o!==t?e.strings[e.strings.length-1]+=o:e.strings.push(""):(e.singleQuote^=1,e.strings[e.strings.length-1]+=o):(e.doubleQuote^=1,e.strings[e.strings.length-1]+=o),e)),{strings:[""],doubleQuote:0,singleQuote:0}).strings}function getParentOrHost(e,t){const o=e.parentNode;return o&&o.host&&11===o.nodeType?o.host:o===t?null:o}function gatherAllElementsXFind(e=null,t){const o=[],n=e=>{for(const t of e)o.push(t),t.shadowRoot&&n(t.shadowRoot.querySelectorAll("*"))};return t.shadowRoot&&n(t.shadowRoot.querySelectorAll("*")),n(t.querySelectorAll("*")),e?o.filter((t=>t.matches(e))):o}function validateSelector(e){try{document.createElement("div").querySelector(e)}catch{throw new Error(`Invalid selector: ${e}`)}}function highlightElements(e){const t=e=>{const t=e.style.outline;e.style.outline="2px solid red",setTimeout((()=>{e.style.outline=t}),2e3)};Array.isArray(e)?e.forEach(t):e&&t(e)}function showWelcomeMessage(){console.log("%cWelcome to Shadow DOM Selector!","color: green; font-size: 16px;"),console.log("%cAuthor: Roland Ross L. Hadi","color: blue; font-size: 14px;"),console.log("%cGitHub: https://github.com/rolandhadi/shadow-dom-selector","color: blue; font-size: 14px;"),console.log("%cExample usage: xfind('downloads-item:nth-child(4) #remove');","color: orange; font-size: 14px;")}showWelcomeMessage();


function onLoadHandler()  {
    "use strict";
    console.log("In onLoadHandler")
    // Bokeh's DocumentReady event is not working, maybe because we work at the Panel level or because it's a standalone doc
    //setTimeout(delayedAdder, 2000);     
    setInterval(delayedAdder, 1000) // keep looking for links created by Panel's Tabulator when scrolling the tables 
}

let lines 
let codeDOM

function delayedAdder() {
    "use strict";
    codeDOM = xfind('pre>code')
    if (codeDOM == null) {
        return
    }
    lines = Array.from(codeDOM.querySelectorAll('a[id^="L"]'))

    //stdoutDOM = xfindAll('.ansi2html-content')
    const anchorlinks = xfindAll('a[href^="#"]')
    anchorlinks.forEach((d) => {
        d.addEventListener("click", clickInterceptor)
    })

    window.addEventListener("popstate", PopStateHandler)
    console.log("delayAdder finished")
}

function clearSourceHighlight() {
    "use strict";
    const highlighted = Array.from(codeDOM.querySelectorAll('.highlighted'))
    highlighted.forEach( h => {
        h.classList.remove("highlighted")
    })
    console.log("Cleared highlights")
}

function clickInterceptor(e) {
    "use strict";
    const target = e.target;
    console.log("clickInterceptor:"+target)
    debugger
    if (target.matches('a[href^="#"]')) {
        e.preventDefault();

        stateData = {
            scrollTop: document.documentElement.scrollTop
        }
        window.history.pushState(stateData,"title",window.location.href)

        const destURL = target.getAttribute('href').slice(1) //remove #
        console.log(`DestURL = ${destURL}`)
        //is it a line range?
        const re = /L(\d+)-(\d+)/
        let range = re.exec(destURL)
        let destID
        if (range !== null) {
            console.log(`range ${range[1]}-${range[2]}`)
            destID = `L${range[1]}`
        } else {
            destID = destURL
        }
        const anchorSelector = 'a[id^="' + destID + '"]' // shouldn't be id= ??
        const dest = xfind(anchorSelector)
        dest.scrollIntoView() // shouldn't be in the IF?
        if (codeDOM.querySelectorAll(anchorSelector) !== null) {
            clearSourceHighlight() // should probably be outside of IF??
            if (range == null) {
                dest.classList.add("highlighted")
            } else {
                for (let i=range[1]; i<=range[2]; i++){
                    const anchorSelector = `a[id="L${i}"]`
                    const dest = xfind(anchorSelector)
                    dest.classList.add("highlighted")
                    console.log(`Highlighted ${anchorSelector}`)
                }
            }
        } else {
            console.log(`Selector not found: ${anchorSelector}`)
        }
        console.log("clickInterceptor done")
    }
}

function PopStateHandler(e) {
    "use strict";
    debugger
    let stateData = e.state;
    console.log("recovering state:" + JSON.stringify(stateData))
    document.documentElement.scrollTop = stateData.scrollTop
}

stateData = {
    scrollTop: 0
};
history.replaceState(stateData, "", document.location.href);

if (document.readyState !== "complete") {
  // Loading hasn't finished yet
  console.log("Still not complete: prepared onLoadHandler")
  window.addEventListener("load", onLoadHandler);
} else {
  // "load" has already fired
  onLoadHandler();
}

</script>
"""

def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('paths', nargs='*', help="File/s to plot. If absent, tries to plot the latest JSON file in the output dir.")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("-p", "--recreate-pickle",action="store_true", help=argparse.SUPPRESS) #unmaintained
    parser.add_argument("-n", "--nbins", default=50)
    #parser.add_argument("-d", "--RCspan", type=int, default=10, help="The span maxRC-minRC (as a % of max) over which a plot is considered interesting")
    parser.add_argument("-x", "--exclude", action='append', default=[], help="DisplayNames matched by this regex will be excluded from the plot")
    parser.add_argument("-o", "--output_dir", default="darum", help="Directory to store the results. Default=%(default)s")
    parser.add_argument("-t", "--top", type=int, default=5, help="Plot only the top N most interesting. Default: %(default)s")
    parser.add_argument("--stop", default=False, action='store_true', help="Generate the HTML file but do not open it.")
    parser.add_argument("-s", "--force-standard-mode", default=False, action='store_true', help="Treat Assertion Batches just like members. Default: autodetect")
    parser.add_argument("-i", "--force-IA-mode", default=False, action='store_true', help="Whether to separate Assertion Batches and focus on them, instead of members. Best for Isolated Assertions mode. Default: autodetect")
    parser.add_argument("-l", "--limitRC", type=Quantity, default=None, help=argparse.SUPPRESS) #help="The RC limit that was used during verification. Used only to check consistency of results. Default: %(default)s")
    parser.add_argument("-b", "--bspan", type=int, default=0, help="A function's histogram will only be plotted if it spans => BSPAN bins. Default: %(default)s")

    args = parser.parse_args()
    return plot(args)

def plot(args) -> int:
    logging.basicConfig() #level=numeric_level,format='%(levelname)s:%(message)s')
    numeric_level = max(logging.DEBUG, logging.WARNING - args.verbose * 10)
    log.setLevel(numeric_level)
    logging.getLogger("darum.log_readers").setLevel(numeric_level)


    if not args.paths:
        # Get the path of the latest JSON file in the output directory
        jsons = glob.glob(args.output_dir+"/*.json")
        if jsons != []:
            latest_json = max(jsons, key=os.path.getmtime)
            print(f"Plotting latest JSON file in output dir: {args.output_dir}/{os.path.basename(latest_json)}")
            args.paths.append(latest_json)
        else:
            sys.exit("Error: No file given, and no JSON file in the output dir.")

    results = readLogs(args.paths, args.recreate_pickle)

    assert len(args.paths) == 1, "Multi-file support is not complete. Please specify only 1 file."
    p = args.paths[0]
    with open(p) as jsonfile:
        darum_context = json.load(jsonfile)['darum']
    sourcecode = {}# = list(darum_context['files'].values())[0].splitlines()
    for fname,fdata in darum_context['files'].items():
        sourcecode[fname] = fdata['contents'].splitlines()
        log.debug(f"Found source for '{fname}'")

    dafny_cmdline: list = darum_context["cmd"]
    try:
        lrc_index = dafny_cmdline.index("--resource-limit")
        args.limitRC = Quantity(dafny_cmdline[lrc_index+1])
        log.debug(f"{args.limitRC=} found in logfile ")
    except:
        pass
    IAmode_in_log = "--isolate-assertions" in dafny_cmdline
    if IAmode_in_log:
        log.debug(f"IAmode setting found in logfile")


    # PROCESS THE DATA
    comment_box = ""
    if args.limitRC is not None:
        comment_box += f"* LimitRC = {args.limitRC}\n"

    # Calculate the max-min span for each DisplayName, and the global maxRC, minRC, minFail
    maxRC = -inf
    minRC = inf
    maxRC_ABs = -inf
    minRC_ABs = inf
    minOoR = inf # min RC of the OoR entries
    minFailures = inf # min RC of the failed entries
    maxFailures = -inf # max RC of the failed entries
    df = pd.DataFrame( columns=["minRC", "maxRC", "slowdown", "success", "OoR","fail","fail_extr","AB","loc_html","loc_txt","diag","displayName", "desc", "src"])
    df.index.name="element"


    filenames : set[str]= set()
    for k,v in results.items():
        f = v.filename
        if len(f)>1:
            filenames.add(f)
    filenames_only_one = len(filenames)==1

    # Digest each entry's list of RCs into a dataframe row of the entry's stats
    for k,v in results.items():
        # for ig in args.exclude:
        #     if ig in k:
        #         log.debug(f"Excluding {k}")
        #         continue
        minRC_entry = min(v.RC, default=inf)
        maxRC_entry = max(v.RC, default=-inf)
        minOoR_entry = min(v.OoR, default=inf)
        minOoR = min(minOoR,minOoR_entry)
        minFailures_entry = min(v.failures, default=inf)
        minFailures = min(minFailures,minFailures_entry)
        maxFailures_entry = max(v.failures, default=-inf)
        maxFailures = max(maxFailures,maxFailures_entry)
        if v.AB==0:
            minRC = min(minRC, minRC_entry)
            maxRC = max(maxRC, maxRC_entry)
        else:
            minRC_ABs = min(minRC_ABs, minRC_entry)
            maxRC_ABs = max(maxRC_ABs, maxRC_entry)
    
        diag = ""

        # if a limit was given, we can do some fine-grained sanity checks
        if args.limitRC is not None and v.AB>0:
            # any RC > limitRC should be in the OoRs, not in the RCs
            # but beware, in IAmode, the vRs' RC is the sum of its ABs, so they can legitimately have RCs > limitRC
            if maxRC_entry > args.limitRC:
                log.warning(f"LimitRC={args.limitRC} but {k}({v.AB=}) has maxRC={Quantity(maxRC_entry)}. Should be OoR! ")
            if minOoR_entry < args.limitRC:
                log.warning(f"MinOoR for {k} is {min(v.OoR)}, should be > LimitRC={args.limitRC}")

        # Calculate the span between max and min
        maxCost_entry = maxRC_entry if len(v.OoR)==0 else minOoR_entry
        slowdown = maxCost_entry/minRC_entry # slowdown
        # info = f"{k:40} {len(v.RC):>10} {smag(minRC_entry):>8}    {smag(maxRC_entry):>6} {span:>8.2%}"
        # log.debug(info)
        fail_extremes = "" if minFailures_entry == inf else f"{smag(minFailures_entry)} - {smag(maxFailures_entry)}"
        loc_txt =  v.loc if filenames_only_one else f"{v.filename}:{v.loc}"
        # if we have the source code for the location, turn the location text into an hyperlink
        src = ""
        if sourcecode.get(v.filename) is None:
            loc_html = loc_txt
        else:
            loc_html = "" if filenames_only_one else f"{v.filename}:"
            loc_range = re.match(r'L(\d+)-(\d+)',v.loc) #Lddd-ddd
            if loc_range:
                loc_html += f'<a href="#{v.loc}">{v.loc}</a>'
            else:
                loc_LC = re.match(r'(\d+):(\d+)',v.loc) #ddd:ddd
                if loc_LC:
                    firstline = int(loc_LC.group(1))
                    col = int(loc_LC.group(2))
                    loc_html += f'<a href="#L{firstline}">{firstline}</a>:{col}'
                    srcline = sourcecode[v.filename][firstline-1]
                    src = srcline.lstrip()
                    leading_whitespace = len(srcline)-len(src) # This is correct if tab == 1 char. Dafny 4.8 does this, hence prints error markers out of place in stdout when there's tabs; some changes might come. https://github.com/dafny-lang/dafny/issues/5718
                    adjusted_col = col-leading_whitespace
                    src = src[:adjusted_col] + '🛑' + src[adjusted_col:] 


        df.loc[k] = {
            "success": len(v.RC),
            "minRC" : minRC_entry,
            "maxRC" : maxRC_entry,
            "slowdown" : slowdown,
            "OoR" : len(v.OoR),
            "fail" : len(v.failures),
            "AB" : v.AB,
            "loc_html"   : loc_html,
            "loc_txt" : loc_txt,
            "diag": diag,
            "displayName": v.displayName,
            "desc": v.description,
            "fail_extr": fail_extremes,
            "src": src
        }

    if minFailures == inf:
        df.drop(columns=["fail_extr"], inplace=True)
    minFailures = Quantity(minFailures)
    minOoR = Quantity(minOoR)
    # assert maxRC < minFail

    df_with_sources = df.loc[df.src!='']
    if df_with_sources.empty:
        df.drop(columns=["src"],inplace=True)

    # Make any per-DN adjustments
    members_with_many_ABs = 0
    DNs = set(df["displayName"].values)
    DNs_number = len(DNs)
    for d in DNs:
        # if a DN only has AB0 and AB1, then copy the location from AB1 to AB0 and drop AB1
        # because AB0 is summarized and easier to detect as non-AB in next steps
        dnABs = df[(df.displayName==d) & (df.AB>1)]
        if dnABs.empty:
            df.loc[(df.displayName==d) & (df.AB==0),"loc_html"] = df.loc[(df.displayName==d) & (df.AB==1),"loc_html"].values[0]
            df.drop(df[(df.displayName==d) & (df.AB==1)].index, inplace=True)
            df.loc[(df.displayName==d),"maxAB"] = 0
        else:
            df.loc[(df.displayName==d),"maxAB"] = max(dnABs.AB)
            members_with_many_ABs += 1


    # At this point, we should have no DNs with only AB1: either was absorbed into AB0 or there's AB2, etc
    assert df.loc[df.maxAB==1].empty, f"Unexpected AB1s: {df.loc[df.maxAB==1]}"

    # Add the emojis
    df.loc[(df.fail>0) & (df.success==0) ,"diag"] += "❌"
    if not df.loc[(df.fail>0)].empty :
        line = f"""Some items had verification failures, whose details aren't stored in the log. Please inspect the <a href="#stdout">stored Dafny stdout</a> further down."""
        log.info(line)
        comment_box += f"* {line}\n"    
    df.loc[(df.OoR>0),"diag"] += "⌛️"
    # An AB that flipflops needs highlighting
    df.loc[((df.fail>0) | (df.OoR>0)) & (df.success>0),"diag"] += "❗️"

    items_only_OoR = (df.OoR>0) & (df.success==0) & (df.fail==0)
    if  items_only_OoR.any() :
        line = f"Some items only had OoR results. Consider verifying again with a higher RC limit."
        log.info(line)
        comment_box += f"* {line}\n"

    # Ranking the items by interestingness is done through a score.
    # A good starting point:
    df["score"] = df.maxRC - df.minRC
    # but there's a lot of corner cases to consider.

    # Items without successes would have NaNs or infs, which breaks the rest of their calculations
    df.loc[~np.isfinite(df.score),"score"] = 0

    # ABs usually have smaller slowdowns and smaller RCs than whole members, so boost them
    AB_boost_factor = 5
    df.loc[df["AB"]>0,"score"] *= AB_boost_factor

    # Big RCs in ABs are even more suspicious than in whole members, so boost them
    # we need a big number. We want it around the max plotted to keep some measure of proportion.
    bigRC = minOoR
    if bigRC == inf: # there were no OoRs!
        bigRC = maxRC
    if bigRC == -inf: # there were no successes??
        bigRC == maxFailures

    # items with only 1 success have slowdown 0, yet a single success between many failures needs highlighting. Boost the score, but tag them
    only1success = (df.success==1) & ((df.fail+df.OoR)>1)
    df.loc[only1success,"score"] = bigRC
    df.loc[only1success,"diag"] += "❓"

    df.loc[df["OoR"]>0,"score"] += bigRC
    df.loc[df["fail"]>0,"score"] += bigRC * 2

    df.sort_values(["score"], ascending=False, kind='stable', inplace=True)

    IAmode_recommended = IAmode_in_log or (members_with_many_ABs > (DNs_number / 2))
    if args.force_standard_mode:
        if IAmode_recommended:
            log.info(f"Setting IAmode OFF, even though most members have multiple ABs")
            comment_box += f"* Isolate-assertions mode forced off\n"
        IAmode = False
    elif args.force_IA_mode:
        if not IAmode_recommended:
            log.info(f"Setting IAmode ON, even though most members have a single AB")
            comment_box += f"* Isolate-assertions mode forced on\n"
        IAmode = True
    else:
        IAmode= IAmode_recommended
        if IAmode_recommended and not IAmode_in_log:
            log.info(f"Setting IAmode ON because most members have multiple ABs")
            comment_box += f"* Most members have multiple ABs, so Isolate-assertions mode was enabled\n"

    # In IA mode, the focus in on the ABs. Separate them.
    if IAmode:
        #put the desc column at the end
        col_list = df.columns.tolist()
        col_list.remove("desc")
        col_list.append("desc")
        df = df[col_list]

        df_vrs = df[df["AB"] == 0]
        df = df[df["AB"]>0]
        log.debug(f"shape of vRs table:{df_vrs.shape}, ABs table:{df.shape}")
    else:
        df_vrs = None
        df.drop(columns="desc",inplace=True)

    # Add a new index that reifies the current order
    df.reset_index(inplace=True)
    df.rename_axis(index="idx",inplace=True)
    # And make it available as an alternative name. Useful for the element names in the plot legend and axis
    df['element_ordered'] = [f"{i} {s}" for i,s in zip(df.index,df["element"])]

    
    if args.limitRC is None:
        if minOoR < inf:
            # There are OoRs. No limitRC was given so we couldn't check while digesting the results. But now we can check a bit.
            # We expect actual results and failures to be smaller than the OoRs.
            mRC = max(maxFailures,maxRC_ABs) if maxFailures < inf else maxRC_ABs
            # if still infinite, means that everything we had was OoRs
            if mRC != inf:
                line = f"Logs contain OoR results, but no limitRC was given. Minimum OoR RC found = {minOoR}."
                log.info(line)
                comment_box += f"* {line}\n"
                OoRstr = f"OoR > {minOoR}"
                limitRC_defacto = minOoR
                if minOoR < maxRC_ABs:
                    line=f"LimitRC must have been <= {minOoR}, yet some results are higher: {maxRC_ABs=}"
                    log.warn(line)
                    comment_box += f"* {line}\n"
                if minOoR < maxFailures:
                    # Dafny bug??
                    line=f"LimitRC must have been <= {minOoR}, yet some failures are higher: {maxFailures=}. **Dafny bug?**"
                    log.warn(line)
                    comment_box += f"* {line}\n"

        else:
            OoRstr = ""
    else:
        limitRC_defacto = args.limitRC
        # we did some checking at the single-result-level while digesting the logs; here we can do global checks
        if args.limitRC < maxRC_ABs:
            line = f"LimitRC={args.limitRC}, yet some results are higher: {maxRC_ABs}"
            log.warn(line)
            comment_box += f"* {line}\n"
        if  maxFailures > args.limitRC:
            line = f"LimitRC={args.limitRC}, yet {maxFailures=} is higher"
            log.info(line)
            comment_box += f"* {line}\n"
        assert args.limitRC <= minOoR, f"LimitRC={args.limitRC}, yet some OoR results are lower: {minOoR=}"
        if minOoR < inf and  minOoR > args.limitRC * 1.1:
            # There are OoRs, but they are suspiciously higher than the given limit.
            line = (f"The min OoR found = {minOoR} is larger than {args.limitRC=}. Dafny/Z3 bug?") #Dafny issue 5525
            log.warn(line)
            comment_box += f"* {line}\n"
        OoRstr = f"OoR > {args.limitRC}"

    failstr: str = OoRstr #"FAILED"# + fstr

    # estimate the worst-case slowdown
    # if any entry has 0 successes then can't estimate
    if not IAmode: # because it's hard to see whether this'd be useful there
        if df.loc[df.success==0].empty:
            sumMaxRC = df.loc[(np.isfinite(df.maxRC)) & (df.OoR==0), "maxRC"].sum()
            sumMaxRC += limitRC_defacto * df.loc[df.OoR>0].shape[0]
            sumMinRC = df.loc[np.isfinite(df.minRC), "minRC"].sum()
            slowdown_total = sumMaxRC/sumMinRC
            line = f"Worst-case total slowdown estimated for this verification: {">" if minOoR < inf else ""}{slowdown_total:.1f}x"
            if not df.loc[df.OoR>0].empty :
                timeouts_pc = (df.loc[df.OoR>0, 'OoR']/(df.loc[df.OoR>0, 'OoR']+df.loc[df.OoR>0, 'success'])).max()*100
                line += f" (+ ~{timeouts_pc:.0f}% timeouts)"
            log.info(line)
            comment_box += f"* {line}\n"
        else:
            line = f"Can't estimate a worst-case slowdown because some member/s failed to verify at all."
            log.info(line)
            comment_box += f"* {line}\n"


    # PREPARATORY CALCULATIONS TO PLOT THE RESULTS

    # When plotting all histograms together, the result distribution might cause some DNs' bars
    # to fall too close together; the plot is not helpful then.
    # So let's remove such histograms.
    # For that, first we need to calculate all the histograms.
    # And for that, we need to decide their bins.
    # So get the min([minRCs]) and max([maxRCs]) of the top candidates.
    df["excluded"] = df["element"].map(lambda dn: dn_is_excluded(dn, args.exclude)).astype(bool)

    minRC_plot = min(df[~df["excluded"]].iloc[0:args.top]["minRC"])
    maxRC_plot = max(df[~df["excluded"]].iloc[0:args.top]["maxRC"])

    plotting_fails = (minOoR < inf) or (minFailures <inf) # whether we'll be plotting failures

    # The histograms have the user-given num of bins between minRC_plot and maxRC_plot,
    # + filler to the left until x=0, + 2 bins if there are fails (margin and fails bar)
    with np.errstate(invalid='ignore'): # silence RuntimeWarnings for inf values
        # inf could be in min/maxRC_plot if all plots are for funcs that failed for all random seeds
        if maxRC_plot < 2*minRC_plot: # the plot would be a big empty space with bars at the right end
            low_bin = 0
        else:
            low_bin = minRC_plot
        bins = np.linspace(Quantity(low_bin),Quantity(maxRC_plot), num=args.nbins+1)
        bin_width = bins[1]-bins[0]
            

    log.debug(f"{args.nbins=}, range {smag(minRC_plot)} - {smag(maxRC_plot)}, bin width {smag(bin_width)}")
    bin_fails_margin = 3 * bin_width
    bin_fails_width = 3 * bin_width
    bin_fails_borders = [bins[-1] + bin_fails_margin, bins[-1] + bin_fails_margin + bin_fails_width]
    bins_with_fails = np.append(bins,bin_fails_borders)
    bin_fails_separator = bins[-1] + 0.5 * bin_fails_margin

    labels_plotted = []
    bins_plot = bins_with_fails if plotting_fails else bins
    bin_centers = 0.5 * (bins_plot[:-1] + bins_plot[1:])
    bin_labels = [smag(b) for b in bin_centers]
    if plotting_fails:
        bin_labels = bin_labels[0:-2] + ["",failstr ]
    hist_df = pd.DataFrame(index = bin_centers)
    plotted = 0
    for i in df.index:
        row = df.loc[i]
        if row.excluded:
            df.loc[i,'diag'] = "⛔️" + df.loc[i,'diag']
            continue
        dnab = row.element
        d = results[dnab]
        nfails = len(d.OoR)+len(d.failures)
        counts, _ = np.histogram(d.RC,bins=bins)
        if plotting_fails:
            counts = np.append(counts,[0,nfails])

        # remove uninteresting plots: those without fails that would span less than <bspan> bins
        nonempty_bins = []
        for b,c in enumerate(counts):
            if c != 0:
                nonempty_bins.append(b)
        bin_span = nonempty_bins[-1]-nonempty_bins[0]

        if (nfails > 0) or (bin_span >= args.bspan):
            labels_plotted.append(dnab)
            hist_df[dnab] = counts
            with np.errstate(divide='ignore'): # to silence the errors because of log of 0 values
                hist_df[dnab+"_log"] = np.log10(counts)
            hist_df[dnab+"_log"] = hist_df[dnab+"_log"].apply(
                lambda l: l if l!=0 else 0.2    # log10(1) = 0, so it's barely visible in plot. log10(2)=0.3. So let's plot 1 as 0.2
                )
            hist_df[dnab+"_RCbin"] = bin_labels # for the hover tool
            df.loc[i,'diag'] = "📊" + df.loc[i,'diag'] #f"F={len(d.failures)} O={len(d.OoR)}"
            plotted += 1
        else:
            #row.diag+=f"{bin_span=}"
            pass
        if plotted >= args.top:
            break

    dropped_cols = ["element_ordered","AB","excluded","displayName","maxAB"]

    dropped_cols_text = dropped_cols.copy()
    dropped_cols_text += ["loc_html","src"]
    print(df.drop(columns=dropped_cols_text, errors='ignore')
            # .rename(columns={
            #     "span"          : "RC span %",
            #     "weighted_span" : "minRC * span"
            #     })
            .head(args.top)
            .to_string (formatters={
                    'maxRC':lambda x: smag(x) if abs(x)!=inf else "-" ,
                    'minRC':lambda x: smag(x) if abs(x)!=inf else "-" ,
                    #'OoRs':smag,
                    #'failures':smag,
                    "slowdown":lambda x: f"{x:>4.2}"
                    },
                na_rep='-',
                float_format=smag
                #max_rows=8
                )
            )

    can_plot = not np.isnan(bin_width) and bin_width >0
    if not can_plot:
        comment_box += f"* The top {args.top} results were not plottable. Only tables were generated.\n"

    print(f"Comments:\n{comment_box}")



    # HOLOVIEWS



    hv.extension('bokeh')
    # renderer = hv.renderer('bokeh')
    # settings.log_level('debug')
    # settings.py_log_level('debug')
    # settings.validation_level('all')

    hvplot = None
    if can_plot:
        histplots_dict = {}
        jitter = (bin_width)/len(labels_plotted)/3
        for i,dn in enumerate(labels_plotted):
            eo = df[df["element"]==dn]["element_ordered"].values[0]
            h = hv.Histogram(
                    (bins_plot+i*jitter,
                        hist_df[dn+"_log"],
                        hist_df[dn],
                        hist_df[dn+"_RCbin"]
                        ),
                    kdims=["RC"],
                    vdims=["LogQuantity", "Quantity", "RCbin"]
                )
            histplots_dict[eo] = h

        hover = HoverTool(tooltips=[
            ("Element", "@Element"),
            ("ResCount bin", "@RCbin"),
            ("Quantity", "@Quantity"),
            ("Log(Quantity)", "@LogQuantity"),
            ])

        bticker = BasicTicker(min_interval = 10**math.floor(math.log10(bin_width)), num_minor_ticks=0)

        hists = hv.NdOverlay(histplots_dict)#, kdims='Elements')
        hists.opts(
            opts.Histogram(alpha=0.9,
                            responsive=True,
                            height=500,
                            tools=[hover],
                            show_legend=True,
                            muted=True,
                            backend_opts={
                            "xaxis.bounds" : (0,bins_plot[-1]+bin_width),
                            "xaxis.ticker" : bticker
                                },
                            autorange='y',
                            ylim=(0,None),
                            xlim=(0,bins_plot[-1]+bin_width),
                            xlabel="RC bins",
                            padding=((0.1,0.1), (0, 0.1)),

                            color=hv.Cycle(),
                ),
            #,logy=True # histograms with logY have been broken in bokeh for years: https://github.com/holoviz/holoviews/issues/2591
            opts.NdOverlay(show_legend=True)
            )

        # A vertical separator for the fails bar
        vline = hv.VLine(bin_fails_separator).opts(
            opts.VLine(color='black', line_width=2, autorange='y',ylim=(0,None),apply_ranges=False)
        )
        vspan = hv.VSpan(bin_fails_separator).opts(
            opts.VSpan(color='#00000020', autorange='y',ylim=(0,None),apply_ranges=False)
        )

        hists = hists * vline * vspan

        ####### SPIKES

        # A JavaScript function to customize the hovertool
        from bokeh.models import CustomJSHover

        RCFfunc = CustomJSHover(code='''
                var value;
                var modified;
                if (value > ''' + str(int(maxRC_plot)) + ''') {
                    modified = "''' + failstr + '''";
                } else {
                    modified = value.toString();
                }
                return modified
        ''')

        nlabs = len(labels_plotted)
        spikes_dict = {}
        for i,dn in enumerate(labels_plotted):
            eo = df[df["element"]==dn]["element_ordered"].values[0]
            RC = results[dn].RC
            # Represent the failures / OoRs with a spike in the last bin
            spike_failures_num = len(results[dn].OoR)+len(results[dn].failures)
            if spike_failures_num > 0:
                spike_failures_sep = bin_width / spike_failures_num
                for f in range(spike_failures_num):
                    RC.append(bin_centers[-1] - bin_width/2 + f*spike_failures_sep)
            hover2 = HoverTool(
                        tooltips=[
                            ("Element", dn),
                            ("ResCount", "@RC{custom}"),
                            ],
                        formatters={
                            "@RC" : RCFfunc,
                            "dn"  : 'numeral'
                        }
                    )
            spikes_dict[eo] = hv.Spikes(RC,kdims="RC").opts(position=nlabs-i-1,tools=[hover2],xaxis="bottom")

        yticks = [(nlabs-i-0.5, list(spikes_dict.keys())[i]) for i in range(nlabs)]#-1,-1,-1)]
        spikes = hv.NdOverlay(spikes_dict).opts(
            yticks = yticks
            )

        spikes.opts(
            opts.Spikes(spike_length=1,
                        line_alpha=1,
                        responsive=True,
                        height=50+nlabs*20,
                        color=hv.Cycle(),
                        ylim=(0,nlabs),
                        autorange=None,
                        yaxis='right',
                        backend_opts={
                            "xaxis.bounds" : (0,bins_plot[-1]+bin_width)
                            },
                        show_legend=False,
                        ),
            opts.NdOverlay(show_legend=False,
                            click_policy='mute',
                            autorange=None,
                            ylim=(0,nlabs),
                            xlim=(0,bins_plot[-1]+bin_width),
                            padding=((0.1,0.1), (0, 0.1)),
                        ),
            #opts.NdOverlay(shared_axes=True, shared_datasource=True,show_legend=False)
            )


        spikes = spikes * vline * vspan


        ##### HISTOGRAMS AND SPIKES TOGETHER

        hvplot = hists + spikes #+ table_plot #+ hist #+ violin
        mf = NumericalTickFormatterWithLimit(bin_fails_separator, format="0.0a")

        hvplot.opts(
        #     #opts.Histogram(responsive=True, height=500, width=1000),
            # opts.Layout(sizing_mode="scale_both", shared_axes=True, sync_legends=True, shared_datasource=True)
            opts.NdOverlay(
                click_policy='mute',
                autorange='y',
                xformatter=mf,
                legend_position="right",
                responsive=True
                )
        )
        hvplot.opts(shared_axes=True)
        hvplot.cols(1)

    # TABLE/S
    dropped_cols += ["loc_txt"]
    #df["span"] = df["span"].apply(lambda d: nan if np.isnan(d) else int(d*10000)/100)
    # We can't use magnitudes with the RCs because then the tables can't be sorted correctly.
    df.minRC = df.minRC.apply(lambda x: x if abs(x)<inf else nan)
    df.maxRC = df.maxRC.apply(lambda x: x if abs(x)<inf else nan)
    df.success = df.success.apply(lambda x: x if x!=0 else nan)
    #df.OoR = df.OoR.apply(lambda x: x if x!=0 else "-")
    #df.fail = df.fail.apply(lambda x: x if x!=0 else "-")

    dft1 = df.drop(columns=dropped_cols).rename(
        columns={
            #"span":"slowdown",
            "loc_html":"location"
            }
    )


    bokeh_formatters = {
        'minRC': NumberFormatter(format='0,0', text_align = 'right',nan_format = '-'),
        'maxRC': NumberFormatter(format='0,0', text_align = 'right'),
        'slowdown': NumberFormatter(format='0.00', text_align = 'right'),
        'score': NumberFormatter(format='0,0', text_align = 'right'),
        'success': NumberFormatter(format='0,0', text_align = 'right'),
        'fail': NumberFormatter(format='0,0', text_align = 'right'),
        'OoR': NumberFormatter(format='0,0', text_align = 'right'),
        'location': {'type':'html'}
    }

    table = pn.widgets.Tabulator(dft1, 
        pagination=None,  # no explicit pages, just scrolling
        frozen_columns=['index'], 
        disabled=True, 
        layout='fit_data_table', 
        selectable=False, 
        text_align={"diag":"center"},
        formatters=bokeh_formatters, 
        configuration={
            "initialSort":[{"column":"score","dir":"asc"}], # not working anyway
            #"renderVertical":"basic" # the whole table is rendered at once, instead of when scrolled in
        }, 
        height=300) #give a glimpse of more rows

    table_title = pn.pane.Markdown("## All elements")
    table_vrs = None
    table_vrs_title = None
    if df_vrs is not None:
        df_vrs.reset_index(inplace=True)
        df_vrs.rename_axis(index="idx",inplace=True)
        #df_vrs["span"] = df_vrs["span"].apply(lambda d: nan if np.isnan(d) else int(d*10000)/100)
        #df_vrs.minRC = df_vrs.minRC.apply(lambda x: x if abs(x)<inf else "-")
        #df_vrs.maxRC = df_vrs.maxRC.apply(lambda x: x if abs(x)<inf else "-")
        #df_vrs.success = df_vrs.success.apply(lambda x: x if x!=0 else "-")
        #df_vrs.OoR = df_vrs.OoR.apply(lambda x: x if x!=0 else "-")
        #df_vrs.fail = df_vrs.fail.apply(lambda x: x if x!=0 else "-")
        dropped_cols += ["diag", "src", "desc"]
        dft2 = df_vrs.drop(columns=dropped_cols, errors='ignore').rename(
                            columns={
                                #"span":"RCspan%",
                                "loc_html":"location"   
                                })

        table_vrs = pn.widgets.Tabulator(dft2, 
            pagination=None,  
            frozen_columns=['index'], 
            disabled=True, 
            layout='fit_data_table', 
            selectable=False, 
            text_align={"diag":"center"},
            formatters=bokeh_formatters, 
            configuration={
                "initialSort":[{"column":"score","dir":"asc"}], # not working anyway
                #"renderVertical":"basic" 
            }, 
            height=300) #give a glimpse of more rows

        table_title = pn.pane.Markdown("## AB-level data")
        table_vrs_title = pn.pane.Markdown("## Member-level summary")


    # the legend and other explanations (score?) could be turned into table tooltips, hoped for Panel 1.5
    legend_icons = """
## Legend
### Elements
MemberName [C] = MemberName's Correctness assertions (as opposed to the default Well-Formedness)  
MemberName B2 = MemberName's Assertion Batch 2  


### Diagnostic icons
❌  All iterations failed verification  
⌛️  Some iteration ran Out of Resources  
❗️  Flipflopping result: some successes, some failures  
❓  Notable entry because there was only 1 success  
📊  Item present in the plot  
⛔️  Item excluded from plot  
"""
    legend_pane = pn.pane.Markdown(legend_icons)

    if comment_box!="":
        comment_box = "# Comments:\n" + comment_box
        pane_comment_box = pn.pane.Markdown(comment_box)
        # print(f"\n{comment_box}")
    else:
        pane_comment_box = None
        

    pane_cmds = pn.Column()
    conv = Ansi2HTMLConverter()
    for p in args.paths:
        try:
            with open(p) as jsonfile:
                j = json.load(jsonfile)['darum']
            pane_cmds.append(pn.pane.Markdown("**" + ' '.join(j['cmd']) + "**"))
            pane_cmds.append(pn.pane.HTML(f"""<a id="stdout"></a>""" + 
                    conv.convert("".join(j['output'])),styles={'background-color': '#CCC'}))
            for fname,fdata in j['files'].items():
    #             source = """Here is an example:

    #     :::python
    #     print('hellow world')
    #     """
    #             source = """Here is an example of AppleScript:
    # ``` { .lang linenos=true linenostart=42 hl_lines="43-44 50" title="An Example Code Block" }
    #     :::python
    #     print('hellow world')
    # ```
    # """
                # The markdown rendereres are supposed to highlight source code. But I can't make them work even to just add line numbers.
                # Pygments doesn't highlight Dafny, anyway.
                # So we add our own line numbers.
                numbered = ""
                splitted = fdata['contents'].splitlines(False)
                lines_max = len(splitted)
                num_digits = int(math.log10(lines_max))
                for i,l in enumerate(splitted):
                    numbered += f'<a id="L{i+1}">{i+1:{num_digits}}: {l}</a><br />'
                stylesheet = '''
a[id^="L"] {
  scroll-margin-top: 50vh;
}
.highlighted {
  background-color : yellow
}
'''
                pane_cmds.append(pn.pane.HTML(f'<h2 id="title">{fname}</h2><pre><code>{numbered}</code></pre>', stylesheets=[stylesheet]))#, renderer="markdown",extensions=["fenced_code","codehilite"]))
        except Exception as e:
            log.info(f"Failed to get extra context data from {p}:{e}")
            continue

    title = "-".join([os.path.splitext(os.path.basename(p))[0] for p in args.paths])
    pane_title = pn.pane.Markdown(f"# Log file: {title}")
    pane_customJS = pn.pane.HTML(customJS, visible=False)
    plot = pn.Column(pane_title, hvplot, table_title, table, table_vrs_title, table_vrs,   pane_comment_box, legend_pane, pane_cmds, pane_customJS)


    # fig.xaxis.bounds = (0,bin_fails)

    plotfilepath: str = os.path.join(args.output_dir, title+".html")

    try:
        os.remove(plotfilepath)
    except:
        pass

    #renderer.save(plot, 'plot')
    # from bokeh.resources import INLINE
    # hv.save(plot, plotfilepath, title=title) #resources='inline')
    #hvplot.show(plot)
    plot.save(plotfilepath,title=title)#, resources=INLINE)

    print(f"Created file {plotfilepath}")

    if args.stop:
        log.debug("Stopping as requested.")
        return(0)
    
    webbrowser.open("file://" + os.path.realpath(plotfilepath), new=2, autoraise=False) 
    #os.system(f"open {plotfilepath}")




    #webbrowser.open('plot.html')

    # ls = hv.link_selections.instance()
    # lplot = ls(plot)
    # hv.save(lplot, 'lplot.html')
    # os.system("open lplot.html")

    return 0


# for easier debugging
if __name__ == "__main__":
    main()




#! python3

import argparse
#from matplotlib import table
import panel as pn
from quantiphy import Quantity
import logging as log
from math import inf, nan
import os
import numpy as np
import pandas as pd
from darum.log_readers import Details, readLogs
from quantiphy import Quantity
from pathlib import Path

from bokeh.models import NumeralTickFormatter, HoverTool
from bokeh.util.compiler import TypeScript
from bokeh.models.widgets.tables import NumberFormatter, ScientificFormatter


def smag(i) -> str:
    return f"{Quantity(i):.3}"

def dn_is_excluded(dn, exclude_list):
    for e in exclude_list:
        if e.lower() in dn.lower():
            return True
    return False

def row_from_Details(v: Details):
    minRC_entry = min(v.RC, default=inf)
#    minRC = min(minRC, minRC_entry)
    maxRC_entry = max(v.RC, default=-inf)
#    maxRC = max(maxRC, maxRC_entry)
    minOoR_entry = min(v.OoR, default=inf)
#    minOoR = min(minOoR,minOoR_entry)
    # minFailures_entry = min(v.failures, default=0)
#    minFailures = min(minFailures,minFailures_entry)

    #comment = ""

    # Calculate the slowdown
    maxCost_entry = maxRC_entry if len(v.OoR)==0 else minOoR_entry
    slowdown = maxCost_entry/minRC_entry #if minRC_entry != 0 else 0
    # info = f"{k:40} {len(v.RC):>10} {smag(minRC_entry):>8}    {smag(maxRC_entry):>6} {slowdown:>8.2%}"
    # log.debug(info)
    return {
        "success": len(v.RC),
        "minRC" : minRC_entry,
        "maxRC" : maxRC_entry,
        "slowdown" : slowdown,
        "OoR" : len(v.OoR),
        "fail" : len(v.failures),
        "AB" : v.AB,
        #"comment": comment
    }


class NumericalTickFormatterWithLimit(NumeralTickFormatter):
    min_fail = 0
    min_OoR = 0

    def __init__(self, min_OoR, min_fail, **kwargs):
        super().__init__(**kwargs)
        assert min_OoR < min_fail
        NumericalTickFormatterWithLimit.min_fail = min_fail
        NumericalTickFormatterWithLimit.min_OoR = min_OoR
        NumericalTickFormatterWithLimit.__implementation__ = TypeScript(
"""
import {NumeralTickFormatter} from 'models/formatters/numeral_tick_formatter'

export class NumericalTickFormatterWithLimit extends NumeralTickFormatter {
    static __name__ = '""" + __name__ + """.NumericalTickFormatterWithLimit'
    MIN_FAIL=""" + str(int(min_fail)) + """
    MIN_OOR=""" + str(int(min_OoR)) + """
    doFormat(ticks: number[], _opts: {loc: number}): string[] {
        const formatted = []
        const ticks2 = super.doFormat(ticks, _opts)
        for (let i = 0; i < ticks.length; i++) {
            if (ticks[i] < this.MIN_OOR) {
                formatted.push(ticks2[i])
            } else if (ticks[i] < this.MIN_FAIL) {
                formatted.push('OoR')
            } else {
                formatted.push('FAIL')
            }
        }
        return formatted
    }
}
""")

def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('path_normal', nargs='+')
    parser.add_argument('-i','--path_IA', nargs='+')
    parser.add_argument("-v", "--verbose", action="count", default=0)
    #parser.add_argument("-p", "--recreate-pickle", default=, action="store_true")
    # parser.add_argument("-n", "--nbins", default=50)
    # #parser.add_argument("-d", "--RCspan", type=int, default=10, help="The span maxRC-minRC (as a % of max) over which a plot is considered interesting")
    # parser.add_argument("-x", "--exclude", action='append', default=[], help="DisplayNames matched by this regex will be excluded from plot")
    # #parser.add_argument("-o", "--only", action='append', default=[], help="Only plot DisplayNames with these substrings")
    parser.add_argument("-t", "--top", type=int, default=20, help="Plot only the top N most interesting")
    # parser.add_argument("-s", "--stop", default=False, action='store_true', help="Process the data but stop before plotting")
    # parser.add_argument("-a", "--IAmode", default=False, action='store_true', help="Isolated Assertions mode. Used only for sanity checking.")
    parser.add_argument("-l", "--limitRC", type=Quantity, default=None, help="The RC limit used during verification. Used only for sanity checking.")
    # parser.add_argument("-b", "--bspan", type=int, default=0, help="The minimum bin span for a histogram to be plotted")
    parser.add_argument("-o", "--output_dir", default="darum", help="Directory to store the results. Default=%(default)s")

    args = parser.parse_args()

    numeric_level = log.WARNING - args.verbose * 10
    log.basicConfig(level=numeric_level,format='%(asctime)s-%(levelname)s:%(message)s',datefmt='%H:%M:%S')

    title = Path(args.path_normal[0]).name + "  +  " + Path(args.path_IA[0]).name

    log.debug(f"logs_normal={args.path_normal}")
    results_normal = readLogs(args.path_normal)#, args.recreate_pickle)
    log.debug(f"logs_IA={args.path_IA}")
    results_IA = readLogs(args.path_IA)#, args.recreate_pickle)

    # PROCESS THE DATA

    ABs_present = False
    vr_past_limitRC = ""
    df_IA = pd.DataFrame( columns=["minRC", "maxRC", "slowdown", "success", "OoR","fail","AB"])
    df_IA.index.name="Element"

    for k,v in results_IA.items():
        if v.AB>0:
            # we want to make sure that these are results from an IA run
            ABs_present = True
            # but we only care for the function-level results
            continue
        df_IA.loc[k] = row_from_Details(v)

    assert ABs_present

    df_IA.drop(columns=["AB"],inplace=True)
    #df_IA["score"] = df_IA["maxRC"] - df_IA["minRC"]
    # df_IA = df_IA.sort_values(["failures","OoRs","score"], ascending=False,kind='stable')
    # df_IA.reset_index(inplace=True)
    # df_IA['Element_ordered'] = [f"{i} {s}" for i,s in zip(df_IA.index,df_IA["Element"])]
    colnames = [e for e in df_IA.columns.values if e != "Element"]
    colnames_IA = [c + " IA" for c in colnames]
    renamer = {c:c_IA for c, c_IA in zip(colnames, colnames_IA)}
    df_IA.rename(columns=renamer, inplace=True)

    df = pd.DataFrame( columns=["minRC", "maxRC", "slowdown", "success", "OoR","fail","AB"])
    df.index.name="Element"
    for k,v in results_normal.items():
        if v.AB!=0:
            continue
        if k not in df_IA.index.values:
            #we only want the same funcs that were contained in the IA file
            continue
        df.loc[k] = row_from_Details(v)
    df.drop(columns=["AB"], inplace=True)

    dfc =pd.concat([df, df_IA], axis=1)
    df = dfc

    AB_boost_factor = 1
    # It might make sense to boost them if we hope that to provide advance warning of trouble.
    # But without a clear reason, given that what we want is to pay attention to brittle verification # in normal mode, then better not to boost.
    df["score"] = df["maxRC"] - df["minRC"] + AB_boost_factor*(df_IA["maxRC IA"] - df_IA["minRC IA"])
    df.loc[~np.isfinite(df.score),"score"] = nan
    df = df.sort_values(["fail", "fail IA", "OoR", "OoR IA","score"], ascending=False,kind='stable')
    df.reset_index(inplace=True)
    df['Element_ordered'] = [f"{i} {s}" for i,s in zip(df.index,df["Element"])]

    maxRC = max(max(df["maxRC"]),max(df["maxRC IA"]))
    minRC = min(min(df["minRC"]),min(df["minRC IA"]))
    RCmargin1 = maxRC * 1.2
    RCOoR = maxRC * 1.4
    RCmargin2 = maxRC * 1.6
    RCfailure = maxRC * 1.8
    sep = 1.001 #separation between spikes/markers faked into the OoR/fail areas



    # HOLOVIEWS

    import holoviews as hv # type: ignore
    # import hvplot
    # from hvplot import hvPlot
    from holoviews import opts
    # from bokeh.models.tickers import FixedTicker, CompositeTicker, BasicTicker

    hv.extension('bokeh')
    renderer = hv.renderer('bokeh')


    ####### SPIKES

    # A JavaScript function to customize the hovertool
    from bokeh.models import CustomJSHover
    RCFfunc = CustomJSHover(code='''
            var value;
            var modified;
            if (value > ''' + str(int(RCmargin2)) + ''') {
                modified = "FAILED";
            } else if (value > ''' + str(int(RCmargin1)) + ''') {
                modified = "OoR";
            } else {
                modified = value.toString();
            }
            return modified
    ''')

    labels_plotted = df["Element"].values[:args.top]
    nlabs = len(labels_plotted)
    spikes_dict = {}
    scatter_dict = {}
    for i,dn in enumerate(labels_plotted):
        eo = df.loc[df["Element"]==dn,"Element_ordered"].values[0]
        RCs_IA = results_IA[dn].RC
        RCs_normal = results_normal[dn].RC
        # Represent the failures / OoRs with a spike/dot at x=RCfailure
        for n in range(0,len(results_IA[dn].OoR)):
            RCs_IA.append(RCOoR*pow(sep, n))
        for n in range(0,len(results_IA[dn].failures)):
            RCs_IA.append(RCfailure*pow(sep, n))
        for n in range(0,len(results_normal[dn].OoR)):
            RCs_normal.append(RCOoR*pow(sep, n))
        for n in range(0,len(results_normal[dn].failures)):
            RCs_normal.append(RCfailure*pow(sep, n))

        hover2 = HoverTool(
                    tooltips=[
                        ("Func", eo + " IA"),
                        ("ResCount", "@RC{custom}"),
                        ],
                    formatters={
                        "@RC" : RCFfunc,
                        #"dn"  : 'numeral'
                        }
                )
        hover3 = HoverTool(
                    tooltips=[
                        ("Func", "@Name"),
                        ("ResCount", "@RC{custom}"),
                        ],
                    formatters={
                        "@RC" : RCFfunc,
                        #"dn"  : 'numeral'
                        }
                )
        spikes_dict[eo] = hv.Spikes(RCs_IA,kdims="RC").opts(position=nlabs-i-1,tools=[hover2],xaxis="bottom", logx = True)
        scatter_dict[eo] = hv.Scatter((
                RCs_normal,
                [nlabs-i-0.5]*len(RCs_normal),
                [eo]*len(RCs_normal)
                ),
            kdims="RC",
            vdims=["y", "Name"]
            ).opts(
                    tools=[hover3],
                    xaxis="bottom",
                    logx = True,
                    marker='x',
                    size=10,
                    show_grid=True,
                    xlabel="Log(RC)",
                    ylabel="Func"
                )

    yticks = [(nlabs-i-0.5, list(spikes_dict.keys())[i]) for i in range(nlabs)]#-1,-1,-1)]
    spikes = hv.NdOverlay(spikes_dict).opts(
        yticks = yticks
        )
    scatter = hv.NdOverlay(scatter_dict).opts(
        yticks = yticks
        )

    spikes.opts(
        opts.Spikes(spike_length=1,
                    #line_alpha=1,
                    responsive=True,
                    height=100+nlabs*20,
                    color=hv.Cycle(),
                    ylim=(0,nlabs+1),
                    #autorange=None,
                    yaxis='right',
                    backend_opts={
                        # "xaxis.bounds" : (0,RCfailure)
                        },
                    title="X = normal mode    | = Isolated Assertions mode",
                    show_legend=False,
                    ),

        opts.NdOverlay(show_legend=False,
                        click_policy='mute',
                        autorange=None,
                        ylim=(0,nlabs+1),
                        #xlim=(3000,RCfailure),
                        padding=(0.05),
                        fontscale=1.5
                    ),
        #opts.NdOverlay(shared_axes=True, shared_datasource=True,show_legend=False)
        )

    scatter.opts(
        opts.Scatter(
                responsive=True,
                height=100+nlabs*20,
                color=hv.Cycle(),
                size=20,
                #ylim=(0,nlabs),
                #autorange=None,
                yaxis='right',
                #backend_opts={
                    #"xaxis.bounds" : (0,bins_plot[-1]+bin_width)
                #    },
                show_legend=False,
                ),
        opts.NdOverlay(show_legend=False,
            click_policy='mute',
            autorange=None,
            ylim=(0,nlabs),
            #xlim=(3000,RCfailure),
            padding=(0.05),
        ),

    )

    vspan = hv.VSpan(RCmargin1).opts(
        opts.VSpan(color='#FF000030',show_legend=False) # transparent red
    )
    vspan = vspan * hv.VSpan(RCmargin2).opts(
        opts.VSpan(color='#FF000030',show_legend=False)
    )

    # TABLE

    df.drop(columns=["Element_ordered"], inplace=True)
    #df["slowdown"] = df["slowdown"].apply(lambda d: nan if np.isnan(d) else int(d*10000)/100)
    #df["slowdown IA"] = df["slowdown IA"].apply(lambda d: nan if np.isnan(d) else int(d*10000)/100)
    df["minRC"] = df["minRC"].apply(lambda x: x if abs(x)<inf else nan)
    df["minRC IA"] = df["minRC IA"].apply(lambda x: x if abs(x)<inf else nan)
    df["maxRC"] = df["maxRC"].apply(lambda x: x if abs(x)<inf else nan)
    df["maxRC IA"] = df["maxRC IA"].apply(lambda x: x if abs(x)<inf else nan)
    df["success"] = df["success"].apply(lambda x: x if x!=0 else nan)
    df["success IA"] = df["success IA"].apply(lambda x: x if x!=0 else nan)
    df["OoR"] = df["OoR"].apply(lambda x: x if x!=0 else nan)
    df["OoR IA"] = df["OoR IA"].apply(lambda x: x if x!=0 else nan)
    df["fail"] = df["fail"].apply(lambda x: x if x!=0 else nan)
    df["fail IA"] = df["fail IA"].apply(lambda x: x if x!=0 else nan)

    df.rename(columns={
            "slowdown":"slowd",
            "slowdown IA":"slowd IA",
            "success": "succs",
            "success IA": "succ IA"
        },inplace=True)

    print(df)

    bokeh_formatters = {
        'minRC': NumberFormatter(format='0,0', text_align = 'right'),
        'maxRC': NumberFormatter(format='0,0', text_align = 'right'),
        'slowd': NumberFormatter(format='0.0000', text_align = 'right'),
        'score': NumberFormatter(format='0,0', text_align = 'right'),
        'succs': NumberFormatter(format='0', text_align = 'right'),
        'fail': NumberFormatter(format='0,0', text_align = 'right'),
        'OoR': NumberFormatter(format='0,0', text_align = 'right'),
    }
    bf_keys = list(bokeh_formatters.keys())
    for k in bf_keys:
        bokeh_formatters[k+" IA"]=bokeh_formatters[k]

    table = pn.widgets.Tabulator(df,
            pagination=None,
            frozen_columns=['index'],
            disabled=True,
            layout='fit_data_table',
            selectable=False,
            text_align={"diag":"center"},
            formatters=bokeh_formatters,
            height=300) #give a glimpse of more rows

    # table = hv.Div("<h2>Normal mode:</h2>").opts(height=50) + table


    hvplot = scatter * spikes * vspan
    # hvplot.cols(1)

    mf = NumericalTickFormatterWithLimit(RCmargin1, RCmargin2, format="0.0a")

    hvplot.opts(
    #     #opts.Histogram(responsive=True, height=500, width=1000),
        # opts.Layout(sizing_mode="scale_both", shared_axes=True, sync_legends=True, shared_datasource=True)
        opts.NdOverlay(
            click_policy='mute',
            autorange='y',
            xformatter=mf,
            #legend_position="right",
            responsive=True,
            show_legend=False,
            )
    )

    pane_title = pn.pane.Markdown(f"# Log files: {title}")
    table_title = pn.pane.Markdown(f"## Comparison normal mode vs IA mode")
    plot = pn.Column(pane_title, hvplot, table_title, table)

    # plot.opts(shared_axes=True)

    # fig = hv.render(plot)
    # #hb = fig.traverse(specs=[hv.plotting.bokeh.Histogram])

    # fig.xaxis.bounds = (0,bin_fails)

    title = title.replace(".json","").replace(" ", "")
    plotfilepath: str = os.path.join(args.output_dir, title+".html")

    try:
        os.remove(plotfilepath)
    except:
        pass

    plot.save(plotfilepath, title=title)


    print(f"Created file {plotfilepath}")
    os.system(f"open {plotfilepath}")

    # Repeat the warning
    # if args.limitRC is None:
    #     if minOoR < inf:
    #         log.warning(f"There are OoR results, but no limitRC was given. Min OoR found = {smag(minOoR)}")


    #webbrowser.open('plot.html')

    # ls = hv.link_selections.instance()
    # lplot = ls(plot)
    # hv.save(lplot, 'lplot.html')
    # os.system("open lplot.html")


# for easier debugging
if __name__ == "__main__":
    main()


#! python3
"""
Runs measure-complexity with random seeds until it finds CSV logs with Resource Usage <min and >max for the given DisplayName.
As of Dafny 4.6 this is obsolete because Dafny got more reliable with iterations and reports each random seed.
"""
import argparse
import csv
from math import inf
from os import error, mkdir, rename
import os
import random
import subprocess as sp
import sys
#from shlex import quote
import time
import logging as log
import enum
from datetime import datetime as dt, timedelta as td
from quantiphy import Quantity
# from yaspin import yaspin, Spinner
# from yaspin.spinners import Spinners

def shell(str, **kwargs):
    """Convenient way to run a shell command and get its exit code, stdout, stderr.."""
    r = sp.run(str, shell=True, capture_output=True, text=True, **kwargs)
    #print(r)
    return r

def smag(i) -> str:
    return f"{Quantity(i):.3}"

def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("dafnyfile")
    parser.add_argument("--displayname", default="")
    parser.add_argument("--min", type=Quantity, default=-inf)
    parser.add_argument("--max", type=Quantity, default=inf)
    parser.add_argument("extra_args", nargs='?', default="")
    parser.add_argument("-v", "--verbose", action="count", default=0)

    args = parser.parse_args()

    # loglevel = "debug"
    # numeric_level = getattr(log, loglevel.upper(), None)
    # if not isinstance(numeric_level, int):
    #     raise ValueError('Invalid log level: %s' % loglevel)
    numeric_level = log.WARNING - args.verbose * 10
    log.basicConfig(level=numeric_level,format='%(levelname)s:%(message)s')


    uint32_max = 4294967295
    random.seed()
    ru_gmin = inf
    seed_gmin = None
    ru_gmax = -inf
    seed_gmax = None
    Quantity.set_prefs(strip_zeros=False)
    iterations = 0
    logs_directory: str = "TestResults/" + dt.now().strftime('%Y%m%d-%H%M%S')
    os.makedirs(logs_directory)

    #with yaspin(Spinner("/-\\|",1000), text="Working...") as spinner:
    #    spinner.hide()
    while True:
        iterations += 1
        # Better not to simply sweep random-seed from 0 because we'd be repeating that sequence at each new run
        random_seed = random.randint(0, uint32_max)
        #argstring = f"--random-seed {random_seed} "
        argstring = f"--boogie /randomSeed:{random_seed}"
        #argstring += "--iterations 1 "
        argstring += args.extra_args
        argstring4filename = (argstring + args.displayname).replace("/","").replace("--","").replace(" ","").replace(":","")
        d = dt.now()
        dstr = d.strftime('%Y%m%d-%H%M%S')
        filename = logs_directory + "/" + dstr + "_" + argstring4filename + ".csv"

        #CLI_line: str = fr"dafny44 measure-complexity --log-format csv\;LogFileName='{filename}' {argstring} {args.dafnyfile}"
        CLI_line: str = fr"dafny44 verify --log-format csv\;LogFileName='{filename}' {argstring} {args.dafnyfile}"
        log.debug("executing: " + CLI_line)
        try:
            res = shell(CLI_line, timeout=10)
        except sp.TimeoutExpired as te:
            log.info(f"Timed out:\n{te}")
            #log.info("executed: " + CLI_line)
            continue

        if res.returncode != 0:
            print(f"returncode={res.returncode}")
            print(f"stdout={res.stdout}")
            print(f"stderr={res.stderr}")
            continue
            #exit(res.returncode)

        if args.displayname == "":
            continue

        with open(filename) as csvfile:
            reader = csv.DictReader(csvfile)
            rows_num = 0
            rus: list[int] = []
            for row in reader:
                rows_num += 1
                log.debug(f"CSV row {rows_num}")
                if args.displayname in row['TestResult.DisplayName'] :
                    # if ru == None:
                    #     ru = int(row['TestResult.ResourceCount'])
                    # else:
                    #     # if we run a single iteration, there should be no multiple rows for a given method/function, right? Though those might happen when something is included or whatever...?
                    #     log.error(f"File {filename} contains multiple rows for {args.displayname}!")
                    #     exit(100)
                    rus.append(int(row['TestResult.ResourceCount']))

        if rus == []:
            log.error("DisplayName not in results!")
            exit(100)

        rus.sort()
        rus_min = rus[0]
        rus_max = rus[-1]

        rus_range_str = f"{smag(rus_min)}"

        if rus_min != rus_max:
            rus_range_str += f"..{smag(rus_max)}"

        if rus_min < ru_gmin:
            filename2 = f"{filename[:-4]}MIN{smag(rus_min)}.csv"
            rename(filename, filename2)
            ru_gmin = rus_min
            seed_gmin = random_seed
        elif rus_max > ru_gmax:
            filename2 = f"{filename[:-4]}MAX{smag(rus_max)}.csv"
            rename(filename, filename2)
            ru_gmax = rus_max
            seed_gmax = random_seed

        status = f"iteration {iterations:4}: seed {random_seed:12} -> {len(rus)}/{rows_num} rows, RU {rus_range_str}\t-"
        status += f"\tmin {smag(ru_gmin)} "
        if ru_gmin <= args.min:
            status += "✅"
        status += f"\tmax {smag(ru_gmax)} "
        if ru_gmax >= args.max:
            status += "✅"

        #spinner.text=line #+ "\n"
        #spinner.write(line)
        log.info(status)

        if ru_gmin <= args.min and ru_gmax >= args.max:
            # TODO generate HTML report
            #spinner.write(line)
            break
        else:
            time.sleep(1)

    print(f"Done: {iterations} iterations. Max RU = {ru_gmax} with seed {seed_gmax}. Min RU = {ru_gmin} with seed {seed_gmin}.")
    return 0


from collections import Counter
import csv
from enum import Enum, auto
import json
import logging
from math import ceil, floor, log10
import re
from datetime import datetime as dt, timedelta as td
import os
import pickle
import sys
from typing import Tuple

from quantiphy import Quantity

log = logging.getLogger(__name__)

def smag(i) -> str:
    return f"{Quantity(i):.3}"

def shortenDisplayName(dn:str) -> str:
    new: str = dn.replace(" (well-formedness)","") # WF is almost everywhere, so take it as default and only mention anything non-WF
    new = new.replace(" (correctness)","[C]")
    return new.strip()

class Details: # gathers the results of multiple iterations of verifying a single AB
    def __init__(self) -> None:
        self.displayName: str = "" # without AB
        self.RC: list[int] = []
        self.OoR: list[int] = []   #OutOfResources
        self.failures: list[int] = []
        #self.RC_max: int     #useful for the table
        #self.RC_min: int
        self.loc: str = ""
        self.filename: str = ""
        self.description: str = ""
        self.AB: int = 0 # in IA mode, there's ABs with a number (vcRs) and non-ABs with num 0 (the vRs that sum the RCs of the vcRs)

type resultsType = dict[str, Details] # displayName_with_AB:Details

def mergeResults(r:resultsType, rNew:resultsType):
    if len(rNew) == 0:
        log.warning(f"no results in rNew")
        exit(1)
    for k in rNew:
        if k in r:
            r[k].RC.extend(rNew[k].RC)
            r[k].OoR.extend(rNew[k].OoR)
            r[k].failures.extend(rNew[k].failures)
        else:
            r[k] = rNew[k]

def check_locations_ABs(locations) -> None:
    # checks across the whole file
    for loc,RStoABs in locations.items():
        if len(RStoABs) == 1:
            continue # only 1 rseed, so nothing to compare to check
        ABs_first = None
        # Ensure that, for each location, the ABs stay the same in every run
        for r,ABs in RStoABs.items():
            if ABs_first == None:
                ABs_first = ABs
                # some locations can accumulate many ABs
                # if len(ABs) > 1:
                #     log.debug(f"{loc} has {len(ABs)} ABs: {ABs}")
                continue
            assert ABs_first == ABs, f"{loc} has changing ABs. Until now it was {ABs_first}. But for rseed {r}, it's {ABs}"

def readCSV(fullpath) -> resultsType:
    """Reads the CSV file into the global usages map"""
    raise NotImplementedError("CSV reading needs fixing")
    # needs to be changed to return a resultsType like readJSON does
    # only reason to use CSV is for Dafny <4.5
    rows = 0
    global results
    with open(fullpath) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            rows += 1
            dn = shortenDisplayName(row['TestResult.DisplayName'])
            rc = row['TestResult.ResourceCount']
            results[dn] = results.get(dn,[]) + [int(rc)]
    log.info(f"{fullpath} :{rows} rows")
    return rows

class SkippingReasons(Enum):
    NONE = auto()
    FAIL = auto()
    OOR = auto()


# there's no JSON schema for the logs. The structure is based on what we've seen experimentally,
# so the reader is rather defensive/paranoic, so that any changes in the format don't cause 
# silent failures.
def readJSON(fullpath: str, paranoid=True) -> resultsType: #tuple[resultsType,dict[int,int]]:
    #reads 1 file (possibly containing multiple verification runs)
    results: resultsType = {}

    with open(fullpath) as jsonfile:
        try:
            j = json.load(jsonfile)
            verificationResults = j["verificationResults"]
        except:
            sys.exit("No verificationResults!")
    log.debug(f"{fullpath}: {len(verificationResults)} verificationResults")
    if len(verificationResults) == 0:
        return results
    
    # A JSON verification log contains a list of verificationResults (vR) objects.
    # Each vR corresponds to a member (function, method...)
    # and contains its Display Name, overall Resource Count, verification outcome and the vcResults (Assertion Batches)
    #   Each AB contains its number (vcNum >= 1), outcome, Resource Count, random seed (strange that it's in the vcR instead of the vR), and a list of assertions
    # Verification iterations can be distinguished because of the random seed.
    # If "isolate assertions", split_here, etc were NOT used, then the AB list in a vR will (usually?) contain only 1 AB
    # Conversely, if those are used, then the AB list in a vR will contain multiple ABs
    # So the 2 extremes are:
    # * standard mode: 1 vR = 1 AB with all the assertions
    # * IAmode: 1 vR = multiple ABs with 1 assertion each.
    # Assertions contain the filename, line, col, and description. They do not contain costs.
    # In the analysis we work with costs, so we need to work at the level of ABs, not assertions. So if there is more than 1 assertion in 1 AB, then we don't try to represent their individual line locations in the plots/tables.
    #
    # How this translates to our output:
    # Dafny by default reports at the vR level, and we keep that spirit.
    # The vR stores the summary info for its ABs. For our results, it gets AB number 0. Doesn't contain locations, so we assign something meaningful from the underlying ABs
    # AB0 doesn't appear in the "element name" of the vR, since the vR is not an AB anyway.
    # We store each AB's info separately. So there's AB0, 1,... n
    # If an AB contains only 1 assertion, we store its filename:line:col. If there's more than 1 assertion, we store filename:L<firstLine>-<lastLine>.

    locations: dict[tuple,dict[int,dict[str,str]]] = {} # relate (file,line,col) to {randomseed:{displayname_AB:description}}; allows to compare results per file position
        # the idea is that a given location, across all randomseeds, should have the same ABs and same results/descriptions
    iteration_costs: dict[int,int] = {}
    for vr in verificationResults:
        shortDN = shortenDisplayName(vr["name"])
        vr_RC = vr["resourceCount"]

        # the rseed is only present in the vcrs, but seems to be constant at the vR level
        # so get it from the first one
        # we won't return it to the user, it's only used for sanity checking and error reporting
        try:
            vr_rseed = vr['vcResults'][0]['randomSeed']
        except:
            # logs made by `dafny verify` contain no randomSeed
            # vr_rseed = None
            sys.exit(f"{shortDN} has no random seed. Maybe this log was created by `dafny verify` instead of `measure-complexity`?")

        iteration_costs[vr_rseed] = iteration_costs.get(vr_rseed, 0) + vr_RC
        det_vr = results.get(shortDN)
        if det_vr is None:
            # first appearance of this shortDN
            det_vr = Details()
            det_vr.AB = 0
            det_vr.displayName = shortDN # they only differ in ABs
        else:
            assert det_vr.AB == 0
            assert det_vr.displayName == shortDN

        if vr["outcome"] == "Correct":
            det_vr.RC.append(vr_RC)
        elif vr["outcome"] == "OutOfResource":
            det_vr.OoR.append(vr_RC)
            #assert vr["outcome"] != "Errors", f"{vr["name"]}, rseed={vr_rseed} has error outcome!"
        elif vr["outcome"] == "Errors":
                #log.info(f"{vr["name"]}, rseed={vr_rseed} has error outcome")
            det_vr.failures.append(vr_RC)
        else:
            sys.exit(f"{shortDN}.outcome == {vr["outcome"]}: unknown case!")

        vcRs = vr['vcResults']

        # #find the filename. The first vcr might have an empty list of assertions (for example in IA mode), so keep trying others.
        # ###filename = None
        # #JSON_order = []
        # for vcr in vcRs:
        #     if filename is None:
        #         try:
        #             asst = vcr['assertions'][0]
        #             filename = asst["filename"] #just for convenience of the log consumer, even though vcRs never have any filename/location
        #             #loc = f"{asst['line']}:{asst['col']}"
        #             #break
        #         except:
        #             pass

        #     #JSON_order.append(vcr['vcNum'])

        # assert filename is not None

        # det.filename = filename
        # results[shortDN] = det

        # The vR doesn't have a location. We'll assign it the majoritary filename from the vcRs after we process them.
        filename = None

        # vcRs are not sorted in the logs. But we need them sorted so that we can skip after a failed one. But, what is their real order? the vcNum one, or the JSON log one? asked in #5862
        # assuming here that the vcNum order is the verification order. Checked with an assert further down.
        vcRs = sorted(vcRs, key=lambda vcR:vcR['vcNum'])

        # We will check that the vr's RC equals the sum of the vcrs' RCs.
        vcrs_RC = []

        ABmax = max([vcr['vcNum'] for vcr in vcRs])
        ABdigits = floor(log10(ABmax)+1) # e.g. log10(99) = 1.x, needs 2 digits
        
        skipping_reason : SkippingReasons = SkippingReasons.NONE
        #reminder: in standard mode there's only 1 vcR per vR
        for vcr in vcRs:
            assert vr_rseed == vcr["randomSeed"], f"rseed mismatch: {vr_rseed} vs {vcr["randomSeed"]} in {shortDN}"

            ABn = vcr['vcNum']
            display_name_AB: str =f"{shortDN} AB{ABn:0{ABdigits}}"

            if skipping_reason != SkippingReasons.NONE:
                # why skip instead of keeping all the information for the log consumer?
                # because we're summarizing for the consumer,
                # so the skipped information must be kept apart from the reliable results.
                if skipping_reason == SkippingReasons.FAIL:
                    # after an AB fails, the situation should be equivalent to "assume False && assert X", so it should always be "valid" - but useless!
                    # So confirm that everything after a "Fail" is "Valid", even though we'll ignore it
                    # If this failed, it'd invalidate the assumption that vcNum order is the verification order (#5862)
                    assert vcr["outcome"] == "Valid", f"Skipping after an AB failed, yet {display_name_AB}=={vcr["outcome"]}"
                elif skipping_reason == SkippingReasons.OOR:
                    if vcr["outcome"] != "Valid":
                        log.info(f"While skipping after an OoR: {display_name_AB} failed, so the OoR should turn out valid?")
                        # would be interesting to keep track of this result and confirm:
                        # * whether the OoR turns out valid
                        # * whether this vcr stays failed.
                continue

            det = results.get(display_name_AB)
            if det is None:
                #first appearance of this display_name_AB, so we'll be initializing the details
                det = Details()
                det.AB = ABn
                det.displayName = shortDN

            # Extract the filename, location and descriptions
            # We don't have a way to know which file was input to Dafny, and different assertions can be located at different files. So we assume that:
            # * the majority of assertions will point to the "main" file
            # * the majority file remains constant across ABs in a vR (trivial in standard mode since there's only 1 AB)
            if len(vcr['assertions'])==0:
                # only seems to happen in each AB1 in IAmode ... until Dafny 4.8 at least
                assert ABn == 1 # no way to know if we're in IAmode (or "close enough" to that )
                if det.loc == "": #first appearance
                    det.filename = "-"
                    det.loc = '-'
                    det.description = '-'
                    log.debug(f"{display_name_AB}\t: zero assertions. To be expected in IAmode.")
                else:
                    assert det.loc == '-'

            elif len(vcr['assertions'])==1:
                asst = vcr['assertions'][0]
                if filename is None:
                    filename = asst['filename']
                else:
                    if filename != asst['filename']:
                        log.info(f"{display_name_AB}\t: location={asst['filename']}:{asst['line']}:{asst['col']}, though we assumed filename={filename}")
                if det.loc == "": #first appearance
                    det.filename = asst['filename']
                    det.loc = f"{asst['line']}:{asst['col']}"
                    det.description = asst['description']
                else:
                    # just double-check that previous appearances with this display_name + AB are consistent
                    assert det.filename == asst['filename']
                    assert det.loc == f"{asst['line']}:{asst['col']}"
                    assert det.description == asst['description']
            else:
                # more than 1 assertion. Store the line range.
                # But first we need to find the majoritary filename in the assertions,
                # so that we can ignore the rest in the calculation of the range of line numbers
                filenames_list = [asst['filename'] for asst in vcr['assertions']]
                filenames_counter = Counter(filenames_list)
                filename_maj = filenames_counter.most_common(1)[0][0]
                if filename is None:
                    filename = filename_maj
                else:
                    assert filename == filename_maj

                if det.loc == "": #first appearance
                    if len(filenames_counter)>1:
                        log.info(f"{display_name_AB}\t: storing only majoritary filename in assertions: {filenames_counter}")
                    det.filename = filename
                    lines = sorted([asst['line'] for asst in vcr['assertions'] if asst["filename"]==filename])
                    lines_str = f"L{lines[0]}"
                    if lines[0]!=lines[-1]:
                        lines_str+=f"-{lines[-1]}"
                    det.loc=lines_str
                    det.description = '*'
                else:
                    assert det.filename == filename

            # store the ABs per locationlocation and the ABs in there to check if they stay consistent
            location_current = (det.filename, display_name_AB, det.loc)
            l = locations.get(location_current,{})
            l2 = l.get(vr_rseed,{})
            l2[ABn]=det.description
            l[vr_rseed]=l2
            locations[location_current] = l

            # store the RCs according to result
            vcr_RC = vcr['resourceCount']

            # Ensure that the AB results make sense vs the vR result
            if vcr["outcome"] == "OutOfResource" :
                assert vr["outcome"] == "OutOfResource", f"{display_name_AB}==OoR, {shortDN}=={vr["outcome"]}: unexpected!"
                det.OoR.append(vcr_RC)
                results[display_name_AB] = det
                if ABmax-ABn >0:
                    log.debug(f"{display_name_AB}==OoR, skipping remaining {ABmax-ABn} ABs in {shortDN}")
                skipping_reason = SkippingReasons.OOR
            elif vcr["outcome"] == "Invalid":
                assert vr["outcome"] == "Errors", f"{display_name_AB}==Invalid, {shortDN}=={vr["outcome"]}: unexpected!"
                det.failures.append(vcr_RC)
                results[display_name_AB] = det
                log.debug(f"{display_name_AB}==Invalid, skipping remaining {ABmax-ABn} ABs in {shortDN}")
                skipping_reason = SkippingReasons.FAIL
            elif vcr["outcome"] == "Valid":
                det.RC.append(vcr_RC)
                results[display_name_AB] = det
                vcrs_RC.append(vcr_RC)
            else:
                sys.exit(f"{display_name_AB}.outcome == {vcr["outcome"]}: unexpected!")

        if skipping_reason is SkippingReasons.NONE: # we reached the end of this vR without fails
            # ensure that the vR cost was coherent with the ABs' sum
            assert sum(vcrs_RC) == vr_RC, f"{shortDN}.RC={vr_RC}, but the sum of the vcrs' RCs is {sum(vcrs_RC)}"
            # ensure that the vR result was reported valid
            assert vr["outcome"] == "Correct", f"{shortDN}=={vr["outcome"]} but all its ABs were Valid!"
        else:
            #log.debug(f"Did not check the sum(vcrs_RC)")
            pass

        # TODO assign to the vr the majoritary filename
        #assert filename != None
        if det_vr.filename == "":
            det_vr.filename = filename
        else:
            assert det_vr.filename == filename
        results[shortDN] = det_vr
        
    if paranoid:
        # the extra checks are actually cheap
        check_locations_ABs(locations)

    cost_max = smag(max(iteration_costs.values()))
    cost_min = smag(min(iteration_costs.values()))
    log.info(f"Iteration costs: {cost_min} to {cost_max}")

    return results #,iteration_costs





def readLogs(paths, read_pickle = False, write_pickle = False) -> resultsType:

    results: resultsType = {}
    files = 0
    # to be un/pickled: [files, results]

    t0 = dt.now()
    picklefilepath = "".join(paths)+"v2.pickle"
    if os.path.isfile(picklefilepath) and read_pickle:
        with open(picklefilepath, 'rb') as pf:
            [files, results] = pickle.load(pf)
        print(f"Loaded pickle: {files} files {(dt.now()-t0)/td(seconds=1)}")
        return results
    else:

        for p in paths:
            # os.walk doesn't accept files, only dirs; so we need to process single files separately
            #log.debug(f"root {p}")
            if os.path.isfile(p):
                ext = os.path.splitext(p)[1]
                if ext == ".json":
                    results_read = readJSON(p)
                elif ext == ".csv":
                    results_read = readCSV(p)
                else:
                    sys.exit(f"Unknown file extension {ext} in {p}")
                mergeResults(results, results_read)
                files += 1
                continue
            files_before_root = files
            for dirpath, dirnames, dirfiles in os.walk(p):
                files_before = files
                for f in dirfiles:
                    if not ".json" in f:
                        continue
                    files +=1
                    fullpath = os.path.join(dirpath, f)
                    log.debug(f"file {files}: {fullpath}")
                    ext = os.path.splitext(fullpath)
                    if ext == ".json":
                        results_read = readJSON(fullpath)
                    elif ext == ".csv":
                        results_read = readCSV(fullpath)
                    else:
                        sys.exit(f"Unknown file format: {fullpath}")
                    mergeResults(results, results_read)

            if files_before_root == files:
                print(f"no files found in {p}")
                exit(1)


        log.info(f"Processed {files} files in {(dt.now()-t0)/td(seconds=1)} sec")

        if write_pickle:
            with open(picklefilepath, "wb") as pf:
                pickle.dump([files, results], pf)
        return results




#! python3
"""
Ease bookkeeping of dafny's measure-complexity runs 
by storing the log file with the args in the filename
"""

import argparse
import hashlib
import json
import os
from pathlib import Path
import shutil
# import subprocess as sp
import sys
import time
import logging
from datetime import datetime as dt, timedelta as td, timezone
import psutil
from quantiphy import Quantity
from typing import NoReturn
from sh import Command
from functools import partial
import webbrowser
#from darum import plot_distribution



def main():
    parser = argparse.ArgumentParser(description="Run dafny's measure-complexity, and store the log with additional context and bookeeping information.")
    parser.add_argument("dafnyfiles", nargs="+", help="The dafny file(s) to verify.")
    parser.add_argument("-e", "--extra_args", default="", help="A quoted string of extra arguments to pass to dafny")
    parser.add_argument("-d", "--dafnyexec", default="dafny", help="The dafny executable")
    parser.add_argument("-r", "--rseed", default=None, help="The random seed. By default is seeded with the current time.")
    parser.add_argument("-m", "--mut", default="10", help="Number of mutations. Default=%(default)s")
    parser.add_argument("-f", "--format", default="json", help=argparse.SUPPRESS) # CVS needs updating
    parser.add_argument("-s", "--filter-symbol", help="Only verify symbols containing this substring.")
    parser.add_argument("-l", "--limitRC", type=Quantity, default=Quantity("10M"), help="The Resource Count limit. Accepts magnitudes (K,M,G...). Default=%(default)s")
    parser.add_argument("-i", "--isolate-assertions",action="store_true")
    parser.add_argument("-c", "--verify-included-files",action="store_true", help="Verify included files")
    parser.add_argument("-z", "--z3-path", help="Dafny's ""solver-path"" argument")
    parser.add_argument("-o", "--output_dir", default="darum", help="Directory to store the results. Default=%(default)s")
    parser.add_argument("-v", "--verbose", action="count", default=0, help="Use multiple times to increase verbosity")
    parser.add_argument("--stop",action="store_true", help="Do not call plot_distribution after verification")

    args = parser.parse_args()

    logging.basicConfig() #level=numeric_level,format='%(levelname)s:%(message)s')
    logger = logging.getLogger(__name__)
    numeric_level = max(logging.DEBUG, logging.WARNING - args.verbose * 10)
    logger.setLevel(numeric_level)

    if args.verify_included_files:
        print("Using --verify-included-files. Beware, only the top file's source will be saved into the augmented log")

    try:
        os.makedirs(args.output_dir)
    except:
        pass
    IAstr = "_IA" if args.isolate_assertions else ""
    VIFstr = "_VIF" if args.verify_included_files else ""
    dafnyfiles_str = ""
    rseed = str(int(time.time())) if args.rseed is None else args.rseed
    source_dict  = {}
    for df in args.dafnyfiles:
        dfbase = os.path.basename(df)
        # with open(df, "rb") as f:
        #     digest = hashlib.file_digest(f, "md5")
        with open(df, "r") as f:
            src= f.read()
        mod_tstamp = os.stat(df).st_mtime
        mod_date = dt.fromtimestamp(mod_tstamp, tz=timezone.utc).isoformat()
        digest = hashlib.md5(bytes(src, encoding="utf-8"))
        hash = digest.hexdigest()
        dfsplit = os.path.splitext(dfbase)
        filenamehash = f"{dfsplit[0]}.H{hash[0:4]}{dfsplit[1]}"
        source_dict[dfbase] = {
            "contents": src,
            "hash": hash,
            "modified": mod_date, #to report changes it's more human-friendly to use the modif_date than a hash
        }
        dafnyfiles_str += f"_{dfsplit[0]}_H{hash[0:4]}"
        # take a snapshot of this input file, adding the same hash piece as in the log
        dfcopy = os.path.join(args.output_dir,filenamehash)
        if not os.path.exists(dfcopy):
            shutil.copy2(df,dfcopy)
    z3_str = f"_{Path(args.z3_path).name}" if args.z3_path else ""
    symbol_str = f"_s{args.filter_symbol}" if args.filter_symbol else ""
    dafnyexec_str= os.path.basename(args.dafnyexec) if args.dafnyexec != "dafny" else ""
    rseed_str = f"_r{args.rseed}" if args.rseed else ""
    extra_args_str = f"_E{args.extra_args}" if args.extra_args else ""
    argstring4filename = f"{dafnyexec_str}{dafnyfiles_str}_M{args.mut}_L{args.limitRC}{rseed_str}{IAstr}{VIFstr}{z3_str}{symbol_str}{extra_args_str}".replace("/","").replace("-","").replace(":","").replace(" ","")
    nowstr = dt.now().strftime('%m%d-%H%M%S')
    logfilename = os.path.join(args.output_dir, nowstr + argstring4filename)
    # for convenience, take another snapshot of each single-input-file with the same full filename as the log
    # if len(args.dafnyfiles)==1:
    #     df= args.dafnyfiles[0]
    #     dfsplit = os.path.splitext(df)
    #     shutil.copy2(df,logfilename+dfsplit[1])
    #log.debug(f"filename={filename}")
    #shell_line = fr"{args.dafnyexec} measure-complexity --log-format csv\;LogFileName='{filename}' {args.extra_args} {args.dafnyfile}"

    arglist = [
        # args.dafnyexec,
        "measure-complexity",
        "--random-seed", rseed,
        "--mutations", args.mut,
        "--log-format", f"{args.format};LogFileName={logfilename}.{args.format}",
        "--resource-limit", str(int(args.limitRC)),
        "--isolate-assertions" if args.isolate_assertions else "",
        "--verify-included-files" if args.verify_included_files else "",
        *(["--solver-path", args.z3_path] if args.z3_path else []),
        *(["--filter-symbol", args.filter_symbol] if args.filter_symbol else []),
        *args.extra_args.split(),
        *args.dafnyfiles
        ]
    logger.debug(f"Executing:{args.dafnyexec} {' '.join(arglist)}")
    # sys.stdout.flush()
    # sys.stderr.flush()
    # os.execvp(args.dafnyexec, arglist )

    #pitfalls: bufsize; blocking,

    import atexit
    def killProc():
        logger.debug("Killing the subprocess' group...")
        dafny_proc.kill_group()
    atexit.register(killProc)

    def process_output(stream, store, line):
        nonlocal mutation_tstamp
        nonlocal mutation_times
        nonlocal output_last_tstamp
        nonlocal counter_error
        nonlocal counter_success
        nonlocal state_error
        nonlocal delta_min_last
        now = dt.now()

        if "Starting verification of mutation" in line or "The total consumed resources are" in line:
            if mutation_tstamp is not None:
                delta = int((now - mutation_tstamp).total_seconds())
                if state_error:
                    counter_error +=1
                else:
                    counter_success += 1
                state_error = False
                l = f"DARUM: mutation took {delta} sec.\t{counter_success} successes, {counter_error} errors\n"
                print(l, end=None)
                store.append(l)
                mutation_times.append(delta)
            mutation_tstamp = now

        if "Error" in line:
            state_error = True
        prefix = f'{dt.now().strftime('%H:%M:%S')}: ' if args.verbose>2 else ""
        stream.write(prefix + line)
        store.append(line)
        output_last_tstamp = now
        delta_min_last = 0

    stdout_lines = []
    # stderr = []

    dafny = Command(args.dafnyexec)

    mutation_tstamp = None
    mutation_times = []
    output_last_tstamp = dt.now()
    counter_error = 0
    counter_success = 0
    state_error = False
    dafny_proc = dafny(arglist,_out=partial(process_output, sys.stdout, stdout_lines), _bg=True, _err_to_out=True, _ok_code=[0,1,2,3,4],_return_cmd=True, _new_session=True)
    # p = sp.Popen(arglist, bufsize=-1, stdout=sp.PIPE, stderr=sp.PIPE, text=True, process_group=0)
    # os.set_blocking(p.stdout.fileno(), False)
    # os.set_blocking(p.stderr.fileno(), False)
    pgid = dafny_proc.pgid 
    logger.debug(f"{pgid=}")

    procs_old = []
    delta_min_last = 0
    while dafny_proc.is_alive():
        procs = []
        for proc in psutil.process_iter(['pid', 'name']):
            try:
                proc_pgid = os.getpgid(proc.info['pid'])
                if pgid == proc_pgid:
                    procs.append(proc)
            except:
                pass
        if procs != procs_old:
            logger.info(f"""Child procs: {[f"{proc.info['pid']}({proc.info['name']})" for proc in procs]}""" )
            procs_old = procs
        delta = int((dt.now() - output_last_tstamp).total_seconds())
        if delta // 60 > delta_min_last:
            delta_min_last = delta // 60
            l = f"DARUM: no dafny output for {delta_min_last} minutes..."
            print(l)
            stdout_lines.append(l)
        time.sleep(1)

    dafny_proc.wait()
    atexit.unregister(killProc)

    exit_code = dafny_proc.exit_code
    logger.debug(f"{pgid=}, {exit_code=}")

    print()
    line = f"DARUM:{mutation_times=}"
    print(line)
    stdout_lines.append(line)
    # if a log file was created, add our own data to it
    if exit_code in [0,2,3,4]:
        with open(f"{logfilename}.{args.format}") as jsonfile:
            try:
                json_data = json.load(jsonfile)
                json_data["verificationResults"]
            except:
                logger.error("No verificationResults!")
        darum_context = {}
        darum_context['files']=source_dict
        darum_context['output']=stdout_lines
        darum_context['cmd']=[args.dafnyexec] + arglist
        json_data["darum"]=darum_context
        with open(f"{logfilename}.{args.format}",mode='w') as jsonfile:
            json.dump(json_data,jsonfile)
        print(f"DARUM:Generated augmented logfile at {logfilename}.{args.format}")

    print("\n-----------------------------------------------------------------------------------\n")

    # Check for leaked Z3 processes
    darum_context = dt.now()
    leaked_procs_old = []
    leaked_procs_found = False
    while True:
        elapsed = int((dt.now()-darum_context).total_seconds())
        leaked_procs = []
        for proc in psutil.process_iter(['pid', 'name']):
            try:
                proc_pgid = os.getpgid(proc.info['pid'])
                if pgid == proc_pgid:
                    leaked_procs.append(proc)
            except:
                pass
        if leaked_procs == []:
            break
        leaked_procs_found = True
        if leaked_procs != leaked_procs_old and elapsed>1:
            for proc in leaked_procs:
                logger.warning(f"Leaked process: {proc.info['name']} PID={proc.info['pid']}")
            leaked_procs_old = leaked_procs
        time.sleep(1)
    if leaked_procs_found and elapsed>1:
        logger.warning(f"Leaked processes finished after {elapsed} secs")

    if (args.stop):# or (exit_code not in [0,1,2,3,4]):
        return exit_code

    pd = Command("plot_distribution")
    pd_args = [
        f"{logfilename}.{args.format}",
        *([f"-{"v"*args.verbose}"] if args.verbose>0 else []),
        *(["--force-IA-mode"] if args.isolate_assertions else []),
        *(["--limitRC", str(args.limitRC)] if args.limitRC is not None else []),
    ]

    print(f"DARUM: continuing to `plot_distribution {' '.join(pd_args)}`")
    #print(pd(pd_args,_err_to_out=True))     #alternatives: _fg=True? or redirect _in=sys.stdin, ...
    pd(pd_args,_fg=True)

    return exit_code #from dafny, not from the plot!
